/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2017 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('account-payment-charges', function () {

  juju.components.AccountPaymentCharges = React.createClass({
    displayName: 'AccountPaymentCharges',

    propTypes: {
      acl: React.PropTypes.object.isRequired,
      addNotification: React.PropTypes.func.isRequired,
      getCharges: React.PropTypes.func.isRequired,
      getReceipt: React.PropTypes.func.isRequired,
      username: React.PropTypes.string.isRequired
    },

    getInitialState: function () {
      this.xhrs = [];
      return {
        charges: null,
        loading: false,
        showPopup: ''
      };
    },

    componentWillMount: function () {
      this._getCharges();
    },

    componentWillUnmount: function () {
      this.xhrs.forEach(xhr => {
        xhr && xhr.abort && xhr.abort();
      });
    },

    /**
      Show or hide the receipt popup.
       @param chargeId {String|Null} The charge to display.
    */
    _togglePopup: function (chargeId = null) {
      this.setState({ showPopup: chargeId });
    },

    /**
      Get the payment details for the user.
       @method _getCharges
    */
    _getCharges: function () {
      this.setState({ loading: true }, () => {
        const username = this.props.username;
        const xhr = this.props.getCharges(username, (error, response) => {
          if (error) {
            const message = 'Could not load the list of charges';
            this.props.addNotification({
              title: message,
              message: `${message}: ${error}`,
              level: 'error'
            });
            console.error(message, error);
            return;
          }
          this.setState({ loading: false, charges: response });
        });
        this.xhrs.push(xhr);
      });
    },

    /**
      Generate the line items for a charge.
       @param lineItems {Array} The list of line items.
      @returns {Object} The markup for the line items.
    */
    _generateLineItems: function (lineItems) {
      if (!lineItems || !lineItems.length) {
        return React.createElement(
          'div',
          { className: 'account-payment-charges__line-items' },
          'There are no items for this charge.'
        );
      }
      const items = lineItems.map((item, i) => {
        return React.createElement(
          'li',
          { className: 'user-profile__list-row twelve-col',
            key: item.name + i },
          React.createElement(
            'div',
            { className: 'three-col no-margin-bottom' },
            item.name
          ),
          React.createElement(
            'div',
            { className: 'three-col no-margin-bottom' },
            item.details
          ),
          React.createElement(
            'div',
            { className: 'three-col no-margin-bottom' },
            item.usage
          ),
          React.createElement(
            'div',
            { className: 'three-col last-col no-margin-bottom' },
            item.price
          )
        );
      });
      return React.createElement(
        'div',
        { className: 'account-payment-charges__line-items' },
        React.createElement(
          'h4',
          null,
          'Charges for:'
        ),
        React.createElement(
          'ul',
          { className: 'user-profile__list twelve-col' },
          React.createElement(
            'li',
            { className: 'user-profile__list-header twelve-col' },
            React.createElement(
              'div',
              { className: 'three-col no-margin-bottom' },
              'Name'
            ),
            React.createElement(
              'div',
              { className: 'three-col no-margin-bottom' },
              'Details'
            ),
            React.createElement(
              'div',
              { className: 'three-col no-margin-bottom' },
              'Usage'
            ),
            React.createElement(
              'div',
              { className: 'three-col last-col no-margin-bottom' },
              'Price'
            )
          ),
          items
        )
      );
    },

    /**
      Generate the list of charges.
       @returns {Object} The markup for the list of charges.
    */
    _generateCharges: function () {
      if (this.state.loading) {
        return React.createElement(juju.components.Spinner, null);
      }
      const charges = this.state.charges;
      if (!charges || !charges.length) {
        return React.createElement(
          'div',
          null,
          'You do not have any charges.'
        );
      }
      let list = charges.map(charge => {
        return React.createElement(
          juju.components.ExpandingRow,
          {
            classes: {
              'twelve-col': true,
              'user-profile__list-row': true
            },
            clickable: true,
            key: charge.id },
          React.createElement(
            'div',
            null,
            React.createElement(
              'div',
              { className: 'two-col no-margin-bottom' },
              charge.statementId
            ),
            React.createElement(
              'div',
              { className: 'two-col no-margin-bottom' },
              React.createElement(juju.components.DateDisplay, {
                date: charge.for,
                relative: true })
            ),
            React.createElement(
              'div',
              { className: 'two-col no-margin-bottom' },
              charge.price / 100,
              ' ',
              charge.currency
            ),
            React.createElement(
              'div',
              { className: 'two-col no-margin-bottom' },
              charge.vat / 100,
              ' ',
              charge.currency
            ),
            React.createElement(
              'div',
              { className: 'two-col last-col no-margin-bottom' },
              (charge.price + charge.vat) / 100,
              ' ',
              charge.currency
            ),
            React.createElement(
              'div',
              { className: 'two-col last-col no-margin-bottom' },
              React.createElement(juju.components.GenericButton, {
                action: this._togglePopup.bind(this, charge.id),
                disabled: false,
                type: 'inline-neutral',
                title: 'Show receipt' })
            )
          ),
          React.createElement(
            'div',
            { className: 'twelve-col' },
            this._generateLineItems(charge.lineItems)
          )
        );
      });
      return React.createElement(
        'ul',
        { className: 'user-profile__list twelve-col' },
        React.createElement(
          'li',
          { className: 'user-profile__list-header twelve-col' },
          React.createElement(
            'div',
            { className: 'two-col no-margin-bottom' },
            'ID'
          ),
          React.createElement(
            'div',
            { className: 'two-col no-margin-bottom' },
            'Date'
          ),
          React.createElement(
            'div',
            { className: 'two-col no-margin-bottom' },
            'Price'
          ),
          React.createElement(
            'div',
            { className: 'two-col no-margin-bottom' },
            'VAT'
          ),
          React.createElement(
            'div',
            { className: 'two-col last-col no-margin-bottom' },
            'Total'
          )
        ),
        list
      );
    },

    /**
      Generate the receipt popup.
       @returns {Object} The terms popup markup.
    */
    _generatePopup: function () {
      const charge = this.state.showPopup;
      if (!charge) {
        return null;
      }
      return React.createElement(juju.components.ReceiptPopup, {
        addNotification: this.props.addNotification,
        close: this._togglePopup,
        chargeId: charge,
        getReceipt: this.props.getReceipt });
    },

    render: function () {
      return React.createElement(
        'div',
        { className: 'account-payment-charges' },
        React.createElement(
          'div',
          { className: 'account__section' },
          React.createElement(
            'h2',
            { className: 'account__title twelve-col' },
            'Charges'
          ),
          this._generateCharges(),
          this._generatePopup()
        )
      );
    }

  });
}, '0.1.0', {
  requires: ['date-display', 'expanding-row', 'generic-button', 'loading-spinner', 'receipt-popup']
});